#include <ctype.h>
#include <stdio.h>

#include "mrk/lexer.h"
#include "mrk/lexer_internal.h"

mrk_err mrk_lexer_init(mrk_lexer **out) {
  MRK_CALLOC(out, 1, sizeof(mrk_lexer));

  return mrk_err_ok;
}

void mrk_lexer_free(mrk_lexer *lexer) {
  if (lexer == NULL) {
    return;
  }

  free(lexer);
}

void mrk_lexer_open(mrk_lexer *lexer, const char *buf, size_t len) {
  lexer->buf.s = buf;
  lexer->buf.len = len;
  lexer->pos.line = 0;
  lexer->pos.buf_index = 0;
  lexer->token.start = 0;
  lexer->token.end = 0;
  lexer->token.emitted = false;
}

bool mrk_lexer_done(const mrk_lexer *lexer) {
  return (lexer->buf.len > 0 && lexer->pos.buf_index == lexer->buf.len) ||
         (lexer->buf.s[lexer->pos.buf_index] == '\0');
}

char mrk_lexer_advance(mrk_lexer *lexer) {
  if (mrk_lexer_done(lexer)) {
    return '\0';
  }

  char c = lexer->buf.s[lexer->pos.buf_index];

  // A newline is still part of the previous line, so if the last character was
  // a newline, we now go to the next line
  if (lexer->buf.s[lexer->pos.buf_index] == '\0') {
    lexer->pos.line++;
    lexer->pos.line_index = 0;
  } else {
    lexer->pos.line_index++;
  }

  lexer->pos.buf_index++;
  lexer->token.end++;

  return c;
}

char mrk_lexer_advance_n(mrk_lexer *lexer, size_t n) {
  while (n > 1) {
    mrk_lexer_advance(lexer);
    n--;
  }

  return mrk_lexer_advance(lexer);
}

void mrk_lexer_advance_eq(mrk_lexer *lexer, char c) {
  while (mrk_lexer_peek(lexer) == c) {
    mrk_lexer_advance(lexer);
  }
}

char mrk_lexer_peek(mrk_lexer *lexer) {
  if (mrk_lexer_done(lexer)) {
    return '\0';
  }

  return lexer->buf.s[lexer->pos.buf_index];
}

void mrk_lexer_advance_word(mrk_lexer *lexer) {
  while (!mrk_is_special_char(mrk_lexer_peek(lexer))) {
    mrk_lexer_advance(lexer);
  }
}

bool mrk_lexer_peek_str(mrk_lexer *lexer, const char *s) {
  bool match = true;

  size_t i = 0;
  while (*s != '\0') {
    // Check whether the lexer would be done before matching the entire string
    bool done_in_n =
        (lexer->buf.len > 0 && lexer->pos.buf_index + i == lexer->buf.len) ||
        (lexer->buf.s[lexer->pos.buf_index + i] == '\0');
    match = !done_in_n && (lexer->buf.s[lexer->pos.buf_index + i] == *s);

    i++;
    s++;
  }

  return match;
}

char mrk_lexer_peek_n(mrk_lexer *lexer, size_t n) {
  // Check whether the lexer would be done in n steps
  bool done_in_n = false;

  for (size_t i = 0; i < n && !done_in_n; i++) {
    done_in_n =
        (lexer->buf.len > 0 && lexer->pos.buf_index + i == lexer->buf.len) ||
        (lexer->buf.s[lexer->pos.buf_index + i] == '\0');
  }

  return done_in_n ? '\0' : lexer->buf.s[lexer->pos.buf_index + n];
}

void mrk_lexer_reset(mrk_lexer *lexer) {
  lexer->token.start = lexer->pos.buf_index;
  lexer->token.end = lexer->pos.buf_index;
  lexer->token.emitted = false;
}

void mrk_lexer_emit(mrk_token *out, mrk_lexer *lexer, mrk_token_type type) {
  out->type = type;
  out->start = lexer->token.start;
  out->end = lexer->token.end;

  lexer->token.emitted = true;
}

void mrk_lexer_lex_special(mrk_token *out, mrk_lexer *lexer) {
  char c = mrk_lexer_advance(lexer);
  switch (c) {
  // All these characters have multiple meanings depending on their location
  // in the file and how many there are, so the lexer can only match them as
  // one or more grouped characters
  case '#':
    mrk_lexer_advance_eq(lexer, c);
    mrk_lexer_emit(out, lexer, mrk_token_type_pounds);
    break;
  case '`':
    mrk_lexer_advance_eq(lexer, c);
    mrk_lexer_emit(out, lexer, mrk_token_type_backticks);
    break;
  case '-':
    mrk_lexer_advance_eq(lexer, c);
    mrk_lexer_emit(out, lexer, mrk_token_type_dashes);
    break;
  case '_':
    mrk_lexer_advance_eq(lexer, c);
    mrk_lexer_emit(out, lexer, mrk_token_type_underscores);
    break;
  case '*':
    mrk_lexer_advance_eq(lexer, c);
    mrk_lexer_emit(out, lexer, mrk_token_type_stars);
    break;
  case '=':
    mrk_lexer_advance_eq(lexer, c);
    mrk_lexer_emit(out, lexer, mrk_token_type_equals);
    break;
  case '\t':
    mrk_lexer_advance_eq(lexer, c);
    mrk_lexer_emit(out, lexer, mrk_token_type_tabs);
    break;
  case '>':
    mrk_lexer_advance_eq(lexer, c);
    mrk_lexer_emit(out, lexer, mrk_token_type_right_angle_brackets);
    break;
  case '!':
    mrk_lexer_emit(out, lexer, mrk_token_type_bang);
    break;
  case '[':
    // Checkboxes for lists are lexed separately to simplify the parser later
    // on
    if (mrk_lexer_peek_str(lexer, " ]")) {
      mrk_lexer_advance_n(lexer, 2);
      mrk_lexer_emit(out, lexer, mrk_token_type_unchecked_box);
    } else if (mrk_lexer_peek_str(lexer, "x]")) {
      mrk_lexer_advance_n(lexer, 2);
      mrk_lexer_emit(out, lexer, mrk_token_type_checked_box);
    } else {
      mrk_lexer_emit(out, lexer, mrk_token_type_left_bracket);
    }
    break;
  case ']':
    mrk_lexer_emit(out, lexer, mrk_token_type_right_bracket);
    break;
  case '(':
    mrk_lexer_emit(out, lexer, mrk_token_type_left_paren);
    break;
  case ')':
    mrk_lexer_emit(out, lexer, mrk_token_type_right_paren);
    break;
  case '\\':
    if (mrk_lexer_peek(lexer) == '\n') {
      mrk_lexer_emit(out, lexer, mrk_token_type_line_break);
    } else {
      mrk_lexer_emit(out, lexer, mrk_token_type_backslash);
    }
    break;
    // Two consecutive newlines constitute a blank line, otherwise they're
    // ignored as whitespace
  case '\n':
    if (mrk_lexer_peek(lexer) == '\n') {
      mrk_lexer_advance(lexer);
      mrk_lexer_emit(out, lexer, mrk_token_type_blank_line);
    } else {
      mrk_lexer_emit(out, lexer, mrk_token_type_newline);
    }
    break;
  case ' ': {
    // Either a double space or a line break
    if (mrk_lexer_peek_str(lexer, " \n")) {
      mrk_lexer_advance_n(lexer, 2);

      mrk_lexer_emit(out, lexer, mrk_token_type_line_break);
    } else {
      mrk_lexer_advance_eq(lexer, ' ');
      mrk_lexer_emit(out, lexer, mrk_token_type_spaces);
    }
  } break;
  default: {
    // Match ordered list headers
    if (isdigit(c)) {
      mrk_lexer_advance(lexer);

      while (isdigit(mrk_lexer_peek(lexer))) {
        mrk_lexer_advance(lexer);
      }

      if (mrk_lexer_peek(lexer) == '.') {
        mrk_lexer_advance(lexer);
        mrk_lexer_emit(out, lexer, mrk_token_type_dotted_number);
      }
      // Doesn't end with a dot, so it's just a word that happens to start
      // with a number
      else {
        mrk_lexer_advance_word(lexer);
        mrk_lexer_emit(out, lexer, mrk_token_type_word);
      }
    }
    // Any other special scenarios we simply parse as a word
    else {
      mrk_lexer_advance_word(lexer);
      mrk_lexer_emit(out, lexer, mrk_token_type_word);
    }
  } break;
  }
}

mrk_lexer_err mrk_lexer_next(mrk_token *out, mrk_lexer *lexer) {
  if (mrk_lexer_done(lexer)) {
    return mrk_lexer_err_done;
  }

  mrk_lexer_reset(lexer);

  while (!lexer->token.emitted && !mrk_lexer_done(lexer)) {
    char c = mrk_lexer_advance(lexer);
    switch (c) {
    // All these characters have multiple meanings depending on their location
    // in the file and how many there are, so the lexer can only match them as
    // one or more grouped characters
    case '#':
      mrk_lexer_advance_eq(lexer, c);
      mrk_lexer_emit(out, lexer, mrk_token_type_pounds);
      break;
    case '`':
      mrk_lexer_advance_eq(lexer, c);
      mrk_lexer_emit(out, lexer, mrk_token_type_backticks);
      break;
    case '-':
      mrk_lexer_advance_eq(lexer, c);
      mrk_lexer_emit(out, lexer, mrk_token_type_dashes);
      break;
    case '_':
      mrk_lexer_advance_eq(lexer, c);
      mrk_lexer_emit(out, lexer, mrk_token_type_underscores);
      break;
    case '*':
      mrk_lexer_advance_eq(lexer, c);
      mrk_lexer_emit(out, lexer, mrk_token_type_stars);
      break;
    case '=':
      mrk_lexer_advance_eq(lexer, c);
      mrk_lexer_emit(out, lexer, mrk_token_type_equals);
      break;
    case '\t':
      mrk_lexer_advance_eq(lexer, c);
      mrk_lexer_emit(out, lexer, mrk_token_type_tabs);
      break;
    case '>':
      mrk_lexer_advance_eq(lexer, c);
      mrk_lexer_emit(out, lexer, mrk_token_type_right_angle_brackets);
      break;
    case '!':
      mrk_lexer_emit(out, lexer, mrk_token_type_bang);
      break;
    case '[':
      // Checkboxes for lists are lexed separately to simplify the parser later
      // on
      if (mrk_lexer_peek_str(lexer, " ]")) {
        mrk_lexer_advance_n(lexer, 2);
        mrk_lexer_emit(out, lexer, mrk_token_type_unchecked_box);
      } else if (mrk_lexer_peek_str(lexer, "x]")) {
        mrk_lexer_advance_n(lexer, 2);
        mrk_lexer_emit(out, lexer, mrk_token_type_checked_box);
      } else {
        mrk_lexer_emit(out, lexer, mrk_token_type_left_bracket);
      }
      break;
    case ']':
      mrk_lexer_emit(out, lexer, mrk_token_type_right_bracket);
      break;
    case '(':
      mrk_lexer_emit(out, lexer, mrk_token_type_left_paren);
      break;
    case ')':
      mrk_lexer_emit(out, lexer, mrk_token_type_right_paren);
      break;
    case '\\':
      if (mrk_lexer_peek(lexer) == '\n') {
        mrk_lexer_emit(out, lexer, mrk_token_type_line_break);
      } else {
        mrk_lexer_emit(out, lexer, mrk_token_type_backslash);
      }
      break;
      // Two consecutive newlines constitute a blank line, otherwise they're
      // ignored as whitespace
    case '\n':
      if (mrk_lexer_peek(lexer) == '\n') {
        mrk_lexer_advance(lexer);
        mrk_lexer_emit(out, lexer, mrk_token_type_blank_line);
      } else {
        mrk_lexer_emit(out, lexer, mrk_token_type_newline);
      }
      break;
    case ' ': {
      // Either a double space or a line break
      if (mrk_lexer_peek_str(lexer, " \n")) {
        mrk_lexer_advance_n(lexer, 2);

        mrk_lexer_emit(out, lexer, mrk_token_type_line_break);
      } else {
        mrk_lexer_advance_eq(lexer, ' ');
        mrk_lexer_emit(out, lexer, mrk_token_type_spaces);
      }
    } break;
    default: {
      // Match ordered list headers
      if (isdigit(c)) {
        mrk_lexer_advance(lexer);

        while (isdigit(mrk_lexer_peek(lexer))) {
          mrk_lexer_advance(lexer);
        }

        if (mrk_lexer_peek(lexer) == '.') {
          mrk_lexer_advance(lexer);
          mrk_lexer_emit(out, lexer, mrk_token_type_dotted_number);
        }
        // Doesn't end with a dot, so it's just a word that happens to start
        // with a number
        else {
          mrk_lexer_advance_word(lexer);
          mrk_lexer_emit(out, lexer, mrk_token_type_word);
        }
      }
      // Any other special scenarios we simply parse as a word
      else {
        mrk_lexer_advance_word(lexer);
        mrk_lexer_emit(out, lexer, mrk_token_type_word);
      }
    } break;
    }
  }

  return lexer->token.emitted ? mrk_lexer_err_ok : mrk_lexer_err_done;
}

size_t mrk_token_len(mrk_token t) { return t.end - t.start; }

size_t mrk_token_lexeme(const char **out, mrk_lexer *lexer, mrk_token t) {
  *out = lexer->buf.s + t.start;

  return t.end - t.start;
}
